# ruff: noqa: E402,F401
from .events import (
    EventList,
    EventListMetaclass,
    EventWrapper,
    SelfEventWrapper,
    contextevent,
    event,
    subclasscontextevent,
    subcontextevent,
    subclassevent,
)

__author__ = "Vizonex"

from .__version__ import __version__
